import pytest

from mojentic.llm.tools.date_resolver import ResolveDateTool

@pytest.fixture
def tool():
    return ResolveDateTool()


def test_resolve_date_without_reference(tool):
    result = tool.run(relative_date_found="next Friday")
    assert result["relative_date"] == "next Friday"
    assert "resolved_date" in result


def test_resolve_date_with_reference(tool):
    reference_date = "2023-10-01"
    result = tool.run(relative_date_found="next Friday", reference_date_in_iso8601=reference_date)
    assert result["relative_date"] == "next Friday"
    assert result["resolved_date"] == "2023-10-06"  # Adjust the expected date based on the reference date



import structlog
from ollama import Client, Options, ChatResponse

from mojentic.llm.gateways.llm_gateway import LLMGateway
from mojentic.llm.gateways.models import LLMToolCall, LLMGatewayResponse
from mojentic.llm.gateways.ollama_messages_adapter import adapt_messages_to_ollama

logger = structlog.get_logger()


class OllamaGateway(LLMGateway):
    """
    This class is a gateway to the Ollama LLM service.

    Args:
        host (str): The Ollama host to connect to. Defaults to "http://localhost:11434"
        headers (dict): The headers to send with the request. Defaults to an empty dict.
    """
    def __init__(self, host="http://localhost:11434", headers={}):
        self.client = Client(host=host, headers=headers)

    def _extract_options_from_args(self, args):
        options = Options(
            temperature=args['temperature'],
            num_ctx=args['num_ctx']
        )
        if args['num_predict'] > 0:
            options.num_predict = args['num_predict']
        return options

    def complete(self, **args) -> LLMGatewayResponse:
        """
        This method completes the LLM request by delegating to the Ollama service.

        Args:
            model (str): The name of the model to use, as appears in `ollama list`
            messages (List[LLMMessage]): A list of messages to send to the LLM.
            object_model (Optional[BaseModel]): The model to use for validating the response.
            tools (List[LLMTool]): A list of tools to use with the LLM. If a tool call is requested, the tool will be called and the output will be included in the response.
            temperature (float): The temperature to use for the response. Defaults to 1.0
            num_ctx (int): The number of context tokens to use. Defaults to 32768.
            num_predict (int): The number of tokens to predict. Defaults to no limit.
        Returns:
            LLMGatewayResponse: The response from the Ollama service.
        """
        logger.info("Delegating to Ollama for completion", **args)

        options = self._extract_options_from_args(args)

        ollama_args = {
            'model': args['model'],
            'messages': adapt_messages_to_ollama(args['messages']),
            'options': options
        }

        if 'object_model' in args and args['object_model'] is not None:
            ollama_args['format'] = args['object_model'].model_json_schema()

        if 'tools' in args and args['tools'] is not None:
            ollama_args['tools'] = [t.descriptor for t in args['tools']]

        response: ChatResponse = self.client.chat(**ollama_args)

        object = None
        tool_calls = []

        if 'object_model' in args:
            try:
                object = args['object_model'].model_validate_json(response.message.content)
            except Exception as e:
                logger.error("Failed to validate model in", error=str(e), response=response.message.content,
                             object_model=args['object_model'])

        if response.message.tool_calls is not None:
            tool_calls = [LLMToolCall(name=t.function.name,
                                      arguments={str(k): str(t.function.arguments[k]) for k in t.function.arguments})
                          for t in response.message.tool_calls]

        return LLMGatewayResponse(
            content=response.message.content,
            object=object,
            tool_calls=tool_calls,
        )



from typing import List, Optional

from pydantic import BaseModel

from mojentic.llm.gateways.models import LLMGatewayResponse, LLMMessage
from mojentic.llm.tools.llm_tool import LLMTool


class LLMGateway:
    """
    This is an abstract class from which specific LLM gateways are derived.

    To create a new gateway, inherit from this class and implement the `complete` method.
    """
    def complete(self,
                 model: str,
                 messages: List[LLMMessage],
                 object_model: Optional[BaseModel] = None,
                 tools: Optional[List[LLMTool]] = None,
                 temperature: float = 1.0,
                 num_ctx: int = 32768,
                 num_predict: int = -1) -> LLMGatewayResponse:
        """
        This method should be implemented by derived classes to complete the LLM request.

        Args:
            model (str): The name of the model to use, as appears in `ollama list`
            messages (List[LLMMessage]): A list of messages to send to the LLM.
            object_model (Optional[BaseModel]): The model to use for validating the response.
            tools (List[LLMTool]): A list of tools to use with the LLM. If a tool call is requested, the tool will be called and the output will be included in the response.
            temperature (float): The temperature to use for the response. Defaults to 1.0
            num_ctx (int): The number of context tokens to use. Defaults to 32768.
            num_predict (int): The number of tokens to predict. Defaults to no limit.
        Returns:
            LLMGatewayResponse: The response from the Ollama service.
        """
        raise NotImplementedError



import json

import pytest

from mojentic.llm.gateways.models import LLMMessage, MessageRole, LLMToolCall
from mojentic.llm.gateways.openai_messages_adapter import adapt_messages_to_openai


def test_simple_system_message():
    messages = [LLMMessage(role=MessageRole.System, content="This is a system message")]

    adapted_messages = adapt_messages_to_openai(messages)

    assert adapted_messages == [
        {
            'role': 'system',
            'content': 'This is a system message'
        }
    ]


def test_simple_user_message():
    messages = [LLMMessage(role=MessageRole.User, content="Hello, how are you?")]

    adapted_messages = adapt_messages_to_openai(messages)

    assert adapted_messages == [
        {
            'role': 'user',
            'content': 'Hello, how are you?'
        }
    ]


def test_simple_assistant_message():
    messages = [LLMMessage(role=MessageRole.Assistant, content="I am fine, thank you!")]

    adapted_messages = adapt_messages_to_openai(messages)

    assert adapted_messages == [
        {
            'role': 'assistant',
            'content': 'I am fine, thank you!'
        }
    ]


@pytest.fixture
def tool_name():
    return "tool_name"


@pytest.fixture
def tool_arguments():
    return {"argument": "value"}


@pytest.fixture
def tool_call(tool_name, tool_arguments):
    return LLMToolCall(
        id="abc",
        name=tool_name,
        arguments=tool_arguments
    )


def test_assistant_message_with_tool_call(tool_name, tool_arguments):
    messages = [LLMMessage(
        role=MessageRole.Assistant,
        content="I am fine, thank you!",
        tool_calls=[LLMToolCall(
            id="abc",
            name=tool_name,
            arguments=tool_arguments
        )]
    )]

    adapted_messages = adapt_messages_to_openai(messages)

    assert adapted_messages == [
        {
            'role': 'assistant',
            'content': 'I am fine, thank you!',
            'tool_calls': [
                {
                    'id': 'abc',
                    'type': 'function',
                    'function': {
                        'name': tool_name,
                        'arguments': json.dumps(tool_arguments)
                    }
                }
            ]
        }
    ]


def test_tool_response_message(tool_call):
    messages = [LLMMessage(role=MessageRole.Tool, content='{"date": "Friday"}', tool_calls=[tool_call])]

    adapted_messages = adapt_messages_to_openai(messages)

    assert adapted_messages == [
        {
            'role': 'tool',
            'content': '{"date": "Friday"}',
            'tool_call_id': tool_call.id
        }
    ]



class LLMTool:
    def run(self, **kwargs):
        raise NotImplementedError

    @property
    def descriptor(self):
        raise NotImplementedError

    def matches(self, name: str):
        return name == self.descriptor["function"]["name"]


from enum import Enum
from typing import Optional, List, Union

from pydantic import BaseModel, Field


class MessageRole(Enum):
    """
    The role of the message in the conversation.
    """
    System = 'system'
    User = 'user'
    Assistant = 'assistant'
    Tool = 'tool'


class LLMToolCall(BaseModel):
    """
    A tool call to be made available to the LLM

    Parameters
    ----------
    id: Optional[str]
    """
    id: Optional[str] = None
    name: str
    arguments: dict[str, str]


class LLMMessage(BaseModel):
    """
    A message to be sent to the LLM. These would accumulate during a chat session with an LLM.

    Properties:

    - role (MessageRole): The role of the message in the conversation.
    - content (Optional[str]): The content of the message.
    - object (Optional[BaseModel]): The object representation of the message.
    - tool_calls (Optional[List[LLMToolCall]]): A list of tool calls to be made available to the LLM.
    """
    role: MessageRole = MessageRole.User
    content: Optional[str] = None
    object: Optional[BaseModel] = None
    tool_calls: Optional[List[LLMToolCall]] = None


class LLMGatewayResponse(BaseModel):
    """
    The response from the LLM gateway, abstracting you from the quirks of a specific LLM.
    """
    content: Optional[Union[str, dict[str, str]]] = Field(None, description="The content of the response.")
    object: Optional[BaseModel] = Field(None, description="Parsed response object")
    tool_calls: List[LLMToolCall] = Field(default_factory=list,
                                          description="List of requested tool calls from the llm.")



import pytest
from unittest.mock import MagicMock

from pydantic import BaseModel

from mojentic.llm.llm_broker import LLMBroker
from mojentic.llm.gateways.models import LLMMessage, MessageRole, LLMGatewayResponse, LLMToolCall


@pytest.fixture
def mock_gateway(mocker):
    return mocker.MagicMock()

@pytest.fixture
def llm_broker(mock_gateway):
    return LLMBroker(model="test-model", gateway=mock_gateway)

class DummyModel(BaseModel):
    pass

def test_generate_simple_message(llm_broker, mock_gateway):
    test_response_content = "I am fine, thank you!"
    messages = [LLMMessage(role=MessageRole.User, content="Hello, how are you?")]
    mock_gateway.complete.return_value = LLMGatewayResponse(content=test_response_content, object=None, tool_calls=[])

    result = llm_broker.generate(messages)

    assert result == test_response_content
    mock_gateway.complete.assert_called_once()

def test_generate_with_tool_call(llm_broker, mock_gateway, mocker):
    messages = [LLMMessage(role=MessageRole.User, content="What is the date on Friday?")]
    tool_call = mocker.create_autospec(LLMToolCall, instance=True)
    tool_call.name = "resolve_date"
    tool_call.arguments = {"date": "Friday"}
    mock_gateway.complete.side_effect = [
        LLMGatewayResponse(content="", object=None, tool_calls=[tool_call]),
        LLMGatewayResponse(content="The date is Friday.", object=None, tool_calls=[])
    ]

    mock_tool = mocker.MagicMock()
    mock_tool.matches.return_value = True
    mock_tool.run.return_value = {"resolved_date": "Friday"}

    result = llm_broker.generate(messages, tools=[mock_tool])

    assert result == "The date is Friday."
    assert mock_gateway.complete.call_count == 2
    mock_tool.run.assert_called_once_with(date="Friday")

def test_generate_object(llm_broker, mock_gateway):
    messages = [LLMMessage(role=MessageRole.User, content="Analyze this text.")]
    mock_object = DummyModel()
    mock_gateway.complete.return_value = LLMGatewayResponse(content="", object=mock_object, tool_calls=[])

    result = llm_broker.generate_object(messages, object_model=MagicMock())

    assert result == mock_object
    mock_gateway.complete.assert_called_once()


import ollama
import structlog

logger = structlog.get_logger()


class EmbeddingsGateway:
    def __init__(self, model: str = "mxbai-embed-large"):
        self.model = model

    def calculate(self, text: str):
        logger.debug("calculate", text=text)
        embed = ollama.embed(model=self.model, input=text)
        return embed.embeddings[0]



import json

import structlog
from openai import OpenAI

from mojentic.llm.gateways.llm_gateway import LLMGateway
from mojentic.llm.gateways.models import LLMToolCall, LLMGatewayResponse
from mojentic.llm.gateways.openai_messages_adapter import adapt_messages_to_openai

logger = structlog.get_logger()


class OpenAIGateway(LLMGateway):
    """
    This class is a gateway to the OpenAI LLM service.

    Args:
        api_key (str): The OpenAI API key to use.
    """
    def __init__(self, api_key):
        self.client = OpenAI(api_key=api_key)

    def complete(self, **args) -> LLMGatewayResponse:
        """
        This method completes the LLM request by delegating to the Ollama service.

        Args:
            model (str): The name of the model to use, as appears in `ollama list`
            messages (List[LLMMessage]): A list of messages to send to the LLM.
            object_model (Optional[BaseModel]): The model to use for validating the response.
            tools (List[LLMTool]): A list of tools to use with the LLM. If a tool call is requested, the tool will be called and the output will be included in the response.
            temperature (float): The temperature to use for the response. Defaults to 1.0
            num_ctx (int): The number of context tokens to use. Defaults to 32768.
            num_predict (int): The number of tokens to predict. Defaults to no limit.
        Returns:
            LLMGatewayResponse: The response from the Ollama service.
        """
        openai_args = {
            'model': args['model'],
            'messages': adapt_messages_to_openai(args['messages']),
        }

        completion = self.client.chat.completions.create

        if 'object_model' in args and args['object_model'] is not None:
            openai_args['response_format'] = args['object_model']
            completion = self.client.beta.chat.completions.parse

        if 'tools' in args and args['tools'] is not None:
            openai_args['tools'] = [t.descriptor for t in args['tools']]

        response = completion(**openai_args)

        object = None
        tool_calls = []

        if 'object_model' in args and args['object_model'] is not None:
            object = response.choices[0].message.parsed

        if response.choices[0].message.tool_calls is not None:
            for t in response.choices[0].message.tool_calls:
                arguments = {}
                args_dict = json.loads(t.function.arguments)
                for k in args_dict:
                    arguments[str(k)] = str(args_dict[k])
                tool_call = LLMToolCall(id=t.id, name=t.function.name, arguments=arguments)
                tool_calls.append(tool_call)

        return LLMGatewayResponse(
            content=response.choices[0].message.content,
            object=object,
            tool_calls=tool_calls,
        )
